{% extends "file.html" %}
{# See https://github.com/bokeh/bokeh/blob/master/bokeh/core/_templates/file.html #}

{% block preamble %}
<link rel="shortcut icon" href="https://docs.dask.org/en/latest/_static/images/favicon.ico" />
{% endblock %}